from enum import Enum
from .broadcast import BroadcastDto


class IbClientCommandType(Enum):
    RESET_START_PRICES = "ResetStartPrices"


class IbClientCommandDto(BroadcastDto):
    command: IbClientCommandType



from enum import Enum
from .broadcast import BroadcastDto


class IbClientEventType(Enum):
    CONNECTED = "Connected"
    DISCONNECTED = "Disconnected"
    RECONNECTED = "Reconnected"
    CYCLE_COMPLETE = "CycleComplete"


class IbClientEventDto(BroadcastDto):
    event: IbClientEventType



from enum import Enum


class Channel(Enum):
    TickerList = "TickerList"
    QualifiedContractList = "QualifiedContractList"
    IbClientEvent = "IbClientEventDto"
    IbClientCommand = "IbClientCommandDto"



import asyncio
import websockets
from jgmd.logging import FreeTextLogger, LogLevel
from pydantic import ValidationError
from typing import Any, Dict
import json
from ..models import (
    SubscriptionDto,
    SubscriptionAction,
)


class WebSocketServer:
    def __init__(self, logger: FreeTextLogger):
        self.logger = logger
        self.channel_subscriptions = {}  # Maps channel names to sets of clients

    async def start(self, host="localhost", port=8765):
        """Start the WebSocket server."""
        server = await websockets.serve(self.handle_client, host, port)
        self.logger.logSuccessful(
            lambda: f"WebSocket server started on ws://{host}:{port}"
        )
        await server.wait_closed()

    async def handle_client(self, websocket):
        self.logger.logSuccessful(
            lambda: f"New client connected: {websocket.remote_address}"
        )

        try:
            async for message in websocket:
                self.logger.logDebug(
                    lambda: f"Received message from {websocket.remote_address}: {message}"
                )
                try:
                    # Determine whether it's a subscription or a broadcast and process accordingly
                    data: Dict = json.loads(message)
                    if "action" in data:
                        # a SubscriptionDto
                        subscriptionDto = SubscriptionDto(**data)
                        await self.handle_subscription(subscriptionDto, websocket)
                    else:
                        # should be a BroadcastDto
                        channel = data.get("channel")
                        await self.handle_broadcast(channel, message, websocket)
                except ValidationError as e:
                    await websocket.send(json.dumps({"error": str(e)}))
        except websockets.exceptions.ConnectionClosed:
            self.logger.logSuccessful(
                lambda: f"Client disconnected: {websocket.remote_address}"
            )
        finally:
            self.remove_client_from_all_channels(websocket)

    async def handle_broadcast(self, channel: str, msg: str, sender):
        """Broadcast a message to all clients subscribed to a specific channel."""
        self.logger.logDebug(lambda: f"Broadcasting message: {msg}")
        # channel = msg.channel
        if channel in self.channel_subscriptions:
            self.logger.logDebug(lambda: f"Broadcasting to channel: {channel}")
            for client in self.channel_subscriptions[channel]:
                try:
                    if client == sender:
                        continue
                    await client.send(msg)
                except websockets.exceptions.ConnectionClosed:
                    self.unsubscribe_client(channel, client)

    async def handle_subscription(self, dto: SubscriptionDto, websocket):
        """Handle subscription and unsubscription requests."""
        if dto.action == SubscriptionAction.SUBSCRIBE.value:
            self.subscribe_client(dto.channel, websocket)
        elif dto.action == SubscriptionAction.UNSUBSCRIBE.value:
            self.unsubscribe_client(dto.channel, websocket)

    def subscribe_client(self, channel: str, websocket):
        """Add a client to a channel subscription."""
        if channel not in self.channel_subscriptions:
            self.channel_subscriptions[channel] = set()
        self.channel_subscriptions[channel].add(websocket)
        self.logger.logDebug(lambda: f"Client subscribed to channel: {channel}")

    def unsubscribe_client(self, channel: str, websocket):
        """Remove a client from a channel subscription."""
        if channel in self.channel_subscriptions:
            self.channel_subscriptions[channel].discard(websocket)
            if not self.channel_subscriptions[channel]:
                del self.channel_subscriptions[channel]
            self.logger.logDebug(lambda: f"Client unsubscribed from channel: {channel}")

    def remove_client_from_all_channels(self, websocket):
        """Remove a client from all channel subscriptions."""
        for channel in list(self.channel_subscriptions.keys()):
            self.unsubscribe_client(channel, websocket)


if __name__ == "__main__":
    logger = FreeTextLogger(
        logDirectory="logs", fileName="websocket_server.log", logLevel=LogLevel.DEBUG
    )
    server = WebSocketServer(logger)
    asyncio.run(server.start("localhost", 8765))



from pydantic import BaseModel
from enum import Enum


class SubscriptionAction(Enum):
    SUBSCRIBE = "Subscribe"
    UNSUBSCRIBE = "Unsubscribe"


class SubscriptionDto(BaseModel):
    action: str
    channel: str



import asyncio
import websockets
from typing import Callable
from jgmd.logging import FreeTextLogger, LogLevel
from jgmd.util import exceptionToStr
from ..models import (
    SubscriptionDto,
    Channel,
    SubscriptionAction,
)


class WebSocketClient:
    def __init__(self, onReceive: Callable[[str], None], logger: FreeTextLogger):
        self.logger = logger
        self.onReceive = onReceive
        self._websocket = None
        self._receive_task = None

    async def subscribe(self, channel: Channel):
        """Subscribe to a channel."""
        await self.send(
            SubscriptionDto(
                action=SubscriptionAction.SUBSCRIBE.value, channel=channel.value
            ).model_dump_json()
        )

    async def connect(self, uri: str):
        """Establish a WebSocket connection and start receiving messages."""
        try:
            self._websocket = await websockets.connect(uri)
        except Exception as e:
            self.logger.logError(
                lambda: f"Error connecting to WebSocket server. Please ensure the server is running. Full error: \n{exceptionToStr(e)}"
            )
            raise
        self.logger.logSuccessful(lambda: "Connected to WebSocket server")
        # Start receiving messages in the background
        self._receive_task = asyncio.create_task(self._receive())

    async def send(self, message: str):
        """Send a message through the WebSocket connection."""
        if self._websocket:
            await self._websocket.send(message)
            self.logger.logDebug(lambda: f"Sent: {message}")
        else:
            self.logger.logError(lambda: "WebSocket not connected. Message not sent.")

    async def close(self):
        """Close the WebSocket connection and cancel the receive task."""
        if self._receive_task:
            self._receive_task.cancel()
            await self._receive_task
            self.logger.logSuccessful(lambda: "Receive task successfully cancelled.")
        if self._websocket:
            await self._websocket.close()
            self.logger.logSuccessful(lambda: "WebSocket connection closed.")

    async def _receive(self):
        """Background task to receive messages."""
        try:
            async for message in self._websocket:
                self.logger.logDebug(lambda: f"Received: {message}")
                if self.onReceive:
                    if asyncio.iscoroutinefunction(self.onReceive):
                        await self.onReceive(message)
                    else:
                        self.onReceive(message)
        except asyncio.CancelledError:
            self.logger.logError(lambda: "Receiving messages task cancelled.")
            raise
        except Exception as e:
            self.logger.logError(lambda: f"Error while receiving messages: {e}")


if __name__ == "__main__":
    import argparse
    from ..models import (
        TickerDto,
        TickerList,
    )

    parser = argparse.ArgumentParser(description="WebSocket client identifier")
    parser.add_argument(
        "--id", type=str, required=True, help="Unique identifier for the client"
    )
    args = parser.parse_args()

    def jbg(message):
        print(f"JBG: {message}")

    async def run(id: str):
        logger = FreeTextLogger(
            logDirectory="logs",
            fileName="websocket_client.log",
            logLevel=LogLevel.DEBUG,
        )
        client = WebSocketClient(onReceive=jbg, logger=logger)
        try:
            await client.connect("ws://localhost:8765")
            client.onReceive = jbg
            count = 0
            direction = 1  # 1 for counting up, -1 for counting down

            await client.subscribe(Channel.TickerList)

            while True:
                print(f"Client {id}: Count is {count}")

                # broadcast a message
                tickerDto = TickerDto(conId=1, symbol=f"AAPL_{id}", last=100.0 + count)
                tickerList = TickerList(tickers=[tickerDto])
                await client.send(tickerList.model_dump_json())
                # Adjust count
                count += direction
                if count == 1_000_000 or count == 0:
                    direction *= -1

                await asyncio.sleep(1)
        except asyncio.CancelledError:
            print("Main task cancelled.")
        finally:
            await client.close()

    try:
        asyncio.run(run(args.id))
    except KeyboardInterrupt:
        print("Program terminated.")



from pydantic import BaseModel, Field
from enum import Enum


class BroadcastDto(BaseModel):
    channel: str = Field(
        default_factory=lambda: None
    )  # Default is None, will be set in __init__

    def __init__(self, *args, **kwargs):
        # Call the parent's __init__ method to properly handle initialization
        super().__init__(*args, **kwargs)

        # Dynamically set msgType to the subclass name
        self.channel = self.__class__.__name__



from typing import Any, List, Set, Optional
from .broadcast import BroadcastDto


class TickerDto(BroadcastDto):
    conId: int
    symbol: Any
    last: float
    startPrice: Optional[float] = None
    pctDeviation: Optional[float] = None


class TickerList(BroadcastDto):
    tickers: List[TickerDto]


class QualifiedContractDto(BroadcastDto):
    conId: int
    symbol: str
    secType: str
    exchange: str
    multiplier: Optional[float] = None
    monthOfContract: Optional[str] = None
    tickSize: Optional[float] = None
    watchlist: Optional[str] = None


class QualifiedContractList(BroadcastDto):
    contracts: List[QualifiedContractDto]
